#include <stdio.h>
#include <stdlib.h>
#include <string.H>

#define INSTALL 1
#define UNINSTALL 2
#define EXIT 0
#define BACK 5

#define REFERENCE_BYTES_LENGTH 4

int error = 0;
int nProcessedFiles = 0;
int nSuccess = 0;

char* metaFileName = "pad00000.meta";
char* metaFileBackupName = "pad00000.meta.backup";

typedef enum {NA,RU,KR,JP} GameRegion;

// File Operations
void copyFile(char* srcFileName, char* cpyFileName);
int deleteFile(char* fileName);
void exitProgram(const char* message);
int fileExists(char* fileName);
FILE* openFile(char* fileName,const char* mode);

// Patcher Operations
long locateBytes(char* bytesToLocate, FILE* metaFile);
void menu(int *menuOption, int *gameVersion);
int replaceBytes(char* referenceBytes, char* newBytes, int replacementLength, FILE* metaFile);
int restoreBackup();
void resultMessage();

int main()
{
    int menuOption = -1;
    int gameVersion = -1;

    while (menuOption != EXIT)
    {
        menu(&menuOption, &gameVersion);

        if (menuOption == INSTALL)
        {
            if(!fileExists(metaFileBackupName))
            {
                printf("\nCreating a backup...");
                copyFile(metaFileName,metaFileBackupName);
                printf("\nBackup Created.\n");
            }

            FILE* metaFile = openFile(metaFileName,"rb+");
            error = 0;
            nProcessedFiles = 0;
            nSuccess = 0;

            switch(gameVersion-1)
            {
                case(NA):
                {
                    replaceBytes("\x2B\xB2\xED\xD1","\x01\x00\x00\x00\x64\x03\x01", 7, metaFile);
                    replaceBytes("\xE2\x10\x33\xEB","\x01\x00\x00\x00\x64\x03\x01", 7, metaFile);
                    replaceBytes("\x1B\x51\x08\xD4","\x01\x00\x00\x00\x64\x03\x01", 7, metaFile);

                    printf("Witch Nude Texture:");
                    replaceBytes("\x3C\xC6\x22\xC2","\x01\x00\x00\x00\x64\x03\x01", 7, metaFile);

                    printf("multiplemodeldesc:");
                    replaceBytes("\x51\xC9\xA3\x25","\x01\x00\x00\x00\x64\x03\x01", 7, metaFile);
                }break;
                case(RU):
                {
                    replaceBytes("\x2B\xB2\xED\xD1","\x01\x00\x00\x00\x6D\x9B\x02", 7, metaFile);
                    replaceBytes("\x1B\x51\x08\xD4","\x01\x00\x00\x00\x6D\x9B\x02", 7, metaFile);
                    replaceBytes("\xE2\x10\x33\xEB","\x01\x00\x00\x00\x6D\x9B\x02", 7, metaFile);

                    printf("Witch Nude Texture:");
                    replaceBytes("\x3C\xC6\x22\xC2","\x01\x00\x00\x00\x6D\x9B\x02", 7, metaFile);

                    printf("multiplemodeldesc:");
                    replaceBytes("\x51\xC9\xA3\x25","\x01\x00\x00\x00\x6D\x9B\x02", 7, metaFile);
                }break;
                case(KR):
                {
                    replaceBytes("\x1B\x51\x08\xD4","\x01\x00\x00\x00\x8C\xEC", 6, metaFile);
                    replaceBytes("\x2B\xB2\xED\xD1","\x01\x00\x00\x00\x8C\xEC", 6, metaFile);
                    replaceBytes("\xE2\x10\x33\xEB","\x01\x00\x00\x00\x8C\xEC", 6, metaFile);

                    printf("Witch Nude Texture:");
                    replaceBytes("\x3C\xC6\x22\xC2","\x01\x00\x00\x00\x8C\xEC", 6, metaFile);

                    printf("multiplemodeldesc:");
                    replaceBytes("\x51\xC9\xA3\x25","\x01\x00\x00\x00\x8C\xEC", 6, metaFile);
                }break;
                case(JP):
                {
                    replaceBytes("\x2B\xB2\xED\xD1","\x01\x00\x00\x00\x7B\x66\x01", 7, metaFile);
                    replaceBytes("\x1B\x51\x08\xD4","\x01\x00\x00\x00\x7B\x66\x01", 7, metaFile);
                    replaceBytes("\xE2\x10\x33\xEB","\x01\x00\x00\x00\x7B\x66\x01", 7, metaFile);

                    printf("Witch Nude Texture:");
                    replaceBytes("\x3C\xC6\x22\xC2","\x01\x00\x00\x00\x7B\x66\x01", 7, metaFile);

                    printf("multiplemodeldesc:");
                    replaceBytes("\x51\xC9\xA3\x25","\x01\x00\x00\x00\x7B\x66\x01", 7, metaFile);
                }break;
                default:;
            }
            resultMessage();
            fclose(metaFile);
        }
        else if (menuOption == UNINSTALL)
        {
            if (restoreBackup())
            {
                printf("\nThe mod was uninstalled successfully!\n\n");
                system("PAUSE");
                printf("\n");
            }
            else
            {
                printf("Backup not found. The uninstallation was unsuccessful.");
            }
        }
    }
    return 0;
}

long locateBytes(char* bytesToLocate, FILE* metaFile)
{
    int matchingBytesCount = 0;

    char buffer = '\0';

    rewind(metaFile);

    while(fread(&buffer,1,1,metaFile))
    {
      if (buffer == bytesToLocate[matchingBytesCount])
      {
          matchingBytesCount++;
      }
      else
      {
          matchingBytesCount = 0;
      }

      if(matchingBytesCount == REFERENCE_BYTES_LENGTH)
      {
           return ftell(metaFile);
      }
    }

    return -1;
}

void menu(int *menuOption, int *gameVersion)
{
    *menuOption = -1;
    *gameVersion = -1;
    int menuMaxOption = 1;
    char input[32];

    while(*menuOption < 0 || *menuOption > menuMaxOption)
    {
        system("cls");
        printf("***********************************************************\n");
        printf("*                                                         *\n");
        printf("*   Black Desert Online Resorepless Nude Mod Patcher      *\n");
        printf("*                                                         *\n");
        printf("***********************************************************\n");
        printf("What do you want to do?\n\n");
        printf("[1] - Apply patch\n");
        if (fileExists(metaFileBackupName))
        {
            menuMaxOption = 2;
            printf("[%d] - Uninstall mod\n", UNINSTALL);
        }
        printf("[%d] - Exit\n", EXIT) ;
        printf("\nEnter your choice: ");
        fgets(input,32,stdin);

        // Converts the read string to int
        *menuOption = atoi(input);

        if(*menuOption < 0 || *menuOption > menuMaxOption)
        {
            printf("\nInvalid option!\n\n");
            system("PAUSE");
        }
    }

    if (*menuOption == UNINSTALL || *menuOption == EXIT)
    {
        return;
    }

    while(*gameVersion < 1 || *gameVersion > BACK)
    {
        printf("\nPlease Select your game version:\n");
        printf("1 - NA/EU\n");
        printf("2 - RU\n");
        printf("3 - KR\n");
        printf("4 - JP\n");
        printf("%d - Back\n", BACK);
        printf("\nEnter your choice: ");
        fgets(input,32,stdin);

        // Converts the read string to int
        *gameVersion = atoi(input);

        if(*gameVersion < 1 || *gameVersion > BACK)
        {
            printf("\nInvalid option!\n\n");
            system("PAUSE");
        }

        if(*gameVersion == BACK)
        {
            printf("\n");
            menu(menuOption, gameVersion);
            return;
        }
    }
}


int replaceBytes(char* referenceBytes, char* newBytes, int replacementLength, FILE* metaFile)
{
    int referenceBytesEndPos = locateBytes(referenceBytes,metaFile);
    int bytesReplacedCount = 0;

    printf("\nPatching...\n");
    if (referenceBytesEndPos == -1)
    {
        printf("Error: The expected bytes were not found in your meta file.\n");
        error = 1;
        nProcessedFiles++;
        return 0;
    }
    else
    {
        // Goes to the beginning where the bytes to replace starts
        fseek(metaFile,referenceBytesEndPos,SEEK_SET);

        while(bytesReplacedCount < replacementLength)
        {
            fwrite(&newBytes[bytesReplacedCount],1,1,metaFile);
            bytesReplacedCount++;
        }

        printf("Success!\n\n");
        nProcessedFiles++;
        nSuccess++;

        return 1;
    }
    nProcessedFiles++;
    return 0;
}

int restoreBackup()
{
    if (fileExists(metaFileBackupName))
    {
        deleteFile(metaFileName);
        rename(metaFileBackupName, metaFileName);
        return 1;
    }
    else
    {
        return 0;
        printf("\nNo backup file to restore.\nYour meta file is probably the original already.\n\n");
        system("PAUSE");
        printf("\n");
    }

}

void resultMessage()
{
     printf("\n--------------- Complete! ----------------\n\n");
     printf("Attempts: %d / Success: %d\n", nProcessedFiles, nSuccess);

    if (error)
    {
        if(nSuccess != 0)
        {
            printf("\nWarning: The Patch was partially applied");
            printf("\n\tSome files failed to patch, but if you saw at least one \"Success!\" you will get part of the mod working\n");
        }
        else
        {
            printf("\nError: All files failed to patch");
            restoreBackup();
        }

        printf("\nPossible causes:\n");
        printf("-You have selected the wrong game region\n");
        printf("-This tool is not compatible with your game anymore.\n");
    }

    printf("\n");
    system("PAUSE");
    printf("\n");

}


void copyFile(char* srcFileName, char* cpyFileName)
{
    char buffer;
    size_t totalBytesRead = 0;
    size_t totalBytesWritten = 0;
    size_t bytesRead = 0;

    FILE* srcFile = openFile(srcFileName,"rb");
    if (srcFile == NULL)
    {
        printf("\nFailed to open the file %s\nMake sure the file is located in the same folder as this program.", srcFileName);
        exitProgram("");
    }

    FILE* cpyFile = fopen(cpyFileName,"wb");
    if (cpyFile == NULL)
    {
        printf("\nFailed to create the %s file \nMove this program and the %s file to your desktop to avoid folder permission problems", cpyFileName, srcFileName);
        fclose(srcFile);
        exitProgram("");
    }
    while((bytesRead = fread(&buffer,1,1,srcFile)))
    {
        totalBytesRead += bytesRead;
        totalBytesWritten += fwrite(&buffer,1,1,cpyFile);
    }

    if (totalBytesRead == 0)
    {
        printf("\nThe file %s is empty or it couldn't be read, please restore it with a backup.", srcFileName);
        fclose(srcFile);
        fclose(cpyFile);
        exitProgram("");
    }
    if(totalBytesRead != totalBytesWritten)
    {
        printf("\nSomething went wrong when making a copy of the %s file - bytes read: %d bytes written: %d", srcFileName, totalBytesRead, totalBytesWritten);
        fclose(srcFile);
        fclose(cpyFile);
        exitProgram("");
    }

    fclose(srcFile);
    fclose(cpyFile);
}

int deleteFile(char* fileName)
{
    if (remove(fileName) == 0)
    {
        return 1;
    }
    else
    {
        printf("Warning: The file %s couldn't be deleted.\n", fileName);
        return 0;
    }
}

void exitProgram(const char* message)
{
    printf("\n%s", message);
    printf("\n\nExiting.\n");
    system("PAUSE");
    exit(EXIT_FAILURE);
}

int fileExists(char* fileName)
{
    FILE* fp = fopen(fileName,"rb");
    if(fp)
    {
        fclose(fp);
        return 1;
    }
    return 0;
}

FILE* openFile(char* fileName,const char* mode)
{
    FILE* fp = fopen(fileName,mode);
    if (fp == NULL)
    {
        printf("\nFailed to open the file %s\nMake sure the file is located in the same folder as this program.", fileName);
        exitProgram("");
    }
    return fp;
}
